@@ -7,6 +7,7 @@
 #include <zend_exceptions.h>
 
 extern zend_class_entry *redis_cluster_exception_ce;
+int le_cluster_slot_cache;
 
 /* Debugging methods/
 static void cluster_dump_nodes(redisCluster *c) {
@@ -60,7 +61,7 @@ static void dump_reply(clusterReply *reply, int indent) {
             smart_string_appendl(&buf, "\"", 1);
             break;
         case TYPE_MULTIBULK:
-            if (reply->elements == (size_t)-1) {
+            if (reply->elements < 0) {
                 smart_string_appendl(&buf, "(nil)", sizeof("(nil)")-1);
             } else {
                 for (i = 0; i < reply->elements; i++) {
@@ -90,7 +91,7 @@ static void dump_reply(clusterReply *reply, int indent) {
 /* Recursively free our reply object.  If free_data is non-zero we'll also free
  * the payload data (strings) themselves.  If not, we just free the structs */
 void cluster_free_reply(clusterReply *reply, int free_data) {
-    int i;
+    long long i;
 
     switch(reply->type) {
         case TYPE_ERR:
@@ -100,10 +101,14 @@ void cluster_free_reply(clusterReply *reply, int free_data) {
                 efree(reply->str);
             break;
         case TYPE_MULTIBULK:
-            for (i = 0; i < reply->elements && reply->element[i]; i++) {
-                cluster_free_reply(reply->element[i], free_data);
+            if (reply->element) {
+                if (reply->elements > 0) {
+                    for (i = 0; i < reply->elements && reply->element[i]; i++) {
+                        cluster_free_reply(reply->element[i], free_data);
+                    }
+                }
+                efree(reply->element);
             }
-            if (reply->element) efree(reply->element);
             break;
         default:
             break;
@@ -111,10 +116,9 @@ void cluster_free_reply(clusterReply *reply, int free_data) {
     efree(reply);
 }
 
-static void
+static int
 cluster_multibulk_resp_recursive(RedisSock *sock, size_t elements,
-                                 clusterReply **element, int status_strings,
-                                 int *err)
+                                 clusterReply **element, int status_strings)
 {
     int i;
     size_t sz;
@@ -127,8 +131,7 @@ cluster_multibulk_resp_recursive(RedisSock *sock, size_t elements,
 
         // Bomb out, flag error condition on a communication failure
         if (redis_read_reply_type(sock, &r->type, &len) < 0) {
-            *err = 1;
-            return;
+            return FAILURE;
         }
 
         /* Set our reply len */
@@ -138,8 +141,7 @@ cluster_multibulk_resp_recursive(RedisSock *sock, size_t elements,
             case TYPE_ERR:
             case TYPE_LINE:
                 if (redis_sock_gets(sock,buf,sizeof(buf),&sz) < 0) {
-                    *err = 1;
-                    return;
+                    return FAILURE;
                 }
                 r->len = (long long)sz;
                 if (status_strings) r->str = estrndup(buf, r->len);
@@ -151,27 +153,24 @@ cluster_multibulk_resp_recursive(RedisSock *sock, size_t elements,
                 if (r->len >= 0) {
                     r->str = redis_sock_read_bulk_reply(sock,r->len);
                     if (!r->str) {
-                        *err = 1;
-                        return;
+                        return FAILURE;
                     }
                 }
                 break;
             case TYPE_MULTIBULK:
-                if (r->len >= 0) {
-                    r->elements = r->len;
-                    if (r->len > 0) {
-                        r->element = ecalloc(r->len,sizeof(clusterReply*));
-                        cluster_multibulk_resp_recursive(sock, r->elements, r->element,
-                            status_strings, err);
+                r->elements = r->len;
+                if (r->elements > 0) {
+                    r->element = ecalloc(r->len, sizeof(*r->element));
+                    if (cluster_multibulk_resp_recursive(sock, r->elements, r->element, status_strings) < 0) {
+                        return FAILURE;
                     }
-                    if (*err) return;
                 }
                 break;
             default:
-                *err = 1;
-                return;
+                return FAILURE;
         }
     }
+    return SUCCESS;
 }
 
 /* Return the socket for a slot and slave index */
@@ -207,16 +206,13 @@ clusterReply *cluster_read_resp(redisCluster *c, int status_strings) {
  * command and consumed the reply type and meta info (length) */
 clusterReply*
 cluster_read_sock_resp(RedisSock *redis_sock, REDIS_REPLY_TYPE type,
-                       char *line_reply, size_t len)
+                       char *line_reply, long long len)
 {
     clusterReply *r;
 
     r = ecalloc(1, sizeof(clusterReply));
     r->type = type;
 
-    // Error flag in case we go recursive
-    int err = 0;
-
     switch(r->type) {
         case TYPE_INT:
             r->integer = len;
@@ -238,38 +234,23 @@ cluster_read_sock_resp(RedisSock *redis_sock, REDIS_REPLY_TYPE type,
             break;
         case TYPE_MULTIBULK:
             r->elements = len;
-            if (len != (size_t)-1) {
+            if (r->elements > 0) {
                 r->element = ecalloc(len, sizeof(clusterReply*));
-                cluster_multibulk_resp_recursive(redis_sock, len, r->element,
-                                                 line_reply != NULL, &err);
+                if (cluster_multibulk_resp_recursive(redis_sock, len, r->element, line_reply != NULL) < 0) {
+                    cluster_free_reply(r, 1);
+                    return NULL;
+                }
             }
             break;
         default:
-            cluster_free_reply(r,1);
+            cluster_free_reply(r, 1);
             return NULL;
     }
 
-    // Free/return null on communication error
-    if (err) {
-        cluster_free_reply(r,1);
-        return NULL;
-    }
-
     // Success, return the reply
     return r;
 }
 
-/* Helper to open connection and send AUTH if necessary */
-static zend_always_inline int
-cluster_sock_open(RedisSock *redis_sock)
-{
-    zend_bool need_auth = (redis_sock->auth && redis_sock->status != REDIS_SOCK_STATUS_CONNECTED);
-    if (!redis_sock_server_open(redis_sock) && (!need_auth || !redis_sock_auth(redis_sock ))) {
-        return SUCCESS;
-    }
-    return FAILURE;
-}
-
 /*
  * Helpers to send various 'control type commands to a specific node, e.g.
  * MULTI, ASKING, READONLY, READWRITE, etc
@@ -544,7 +525,7 @@ unsigned short cluster_hash_key(const char *key, int len) {
     // Hash the whole key if we don't find a tailing } or if {} is empty
     if (e == len || e == s+1) return crc16(key, len) & REDIS_CLUSTER_MOD;
 
-    // Hash just the bit betweeen { and }
+    // Hash just the bit between { and }
     return crc16((char*)key+s+1,e-s-1) & REDIS_CLUSTER_MOD;
 }
 
@@ -573,7 +554,7 @@ unsigned short cluster_hash_key_zval(zval *z_key) {
             klen = Z_STRLEN_P(z_key);
             break;
         case IS_LONG:
-            klen = snprintf(buf,sizeof(buf),"%ld",Z_LVAL_P(z_key));
+            klen = snprintf(buf,sizeof(buf),ZEND_LONG_FMT,Z_LVAL_P(z_key));
             kptr = (const char *)buf;
             break;
         case IS_DOUBLE:
@@ -655,12 +636,14 @@ cluster_node_create(redisCluster *c, char *host, size_t host_len,
     zend_llist_init(&node->slots, sizeof(redisSlotRange), NULL, 0);
 
     // Attach socket
-    node->sock = redis_sock_create(host, host_len, port, c->timeout,
-        c->read_timeout, c->persistent, NULL, 0);
+    node->sock = redis_sock_create(host, host_len, port,
+                                   c->flags->timeout, c->flags->read_timeout,
+                                   c->flags->persistent, NULL, 0);
 
-    if (c->auth) {
-        node->sock->auth = zend_string_copy(c->auth);
-    }
+    /* Stream context */
+    node->sock->stream_ctx = c->flags->stream_ctx;
+
+    redis_sock_set_auth(node->sock, c->flags->user, c->flags->pass);
 
     return node;
 }
@@ -841,13 +824,12 @@ PHP_REDIS_API redisCluster *cluster_create(double timeout, double read_timeout,
 
     /* Initialize flags and settings */
     c->flags = ecalloc(1, sizeof(RedisSock));
+    c->flags->timeout = timeout;
+    c->flags->read_timeout = read_timeout;
+    c->flags->persistent = persistent;
     c->subscribed_slot = -1;
     c->clusterdown = 0;
-    c->timeout = timeout;
-    c->read_timeout = read_timeout;
     c->failover = failover;
-    c->persistent = persistent;
-    c->auth = NULL;
     c->err = NULL;
 
     /* Set up our waitms based on timeout */
@@ -872,6 +854,8 @@ cluster_free(redisCluster *c, int free_ctx)
 
     /* Free any allocated prefix */
     if (c->flags->prefix) zend_string_release(c->flags->prefix);
+
+    redis_sock_free_auth(c->flags);
     efree(c->flags);
 
     /* Call hash table destructors */
@@ -882,18 +866,17 @@ cluster_free(redisCluster *c, int free_ctx)
     efree(c->seeds);
     efree(c->nodes);
 
-    /* Free auth info we've got */
-    if (c->auth) zend_string_release(c->auth);
-
     /* Free any error we've got */
     if (c->err) zend_string_release(c->err);
 
-    /* Invalidate our cache if we were redirected during operation */
     if (c->cache_key) {
+        /* Invalidate persistent cache if the cluster has changed */
         if (c->redirections) {
             zend_hash_del(&EG(persistent_list), c->cache_key);
-            c->cache_key = NULL;
         }
+
+        /* Release our hold on the cache key */
+        zend_string_release(c->cache_key);
     }
 
     /* Free structure itself */
@@ -944,34 +927,6 @@ redisCachedCluster *cluster_cache_create(zend_string *hash, HashTable *nodes) {
     return cc;
 }
 
-/* Takes our input hash table and returns a straigt C array with elements,
- * which have been randomized.  The return value needs to be freed. */
-static zval **cluster_shuffle_seeds(HashTable *seeds, int *len) {
-    zval **z_seeds, *z_ele;
-    int *map, i, count, index = 0;
-
-    /* How many */
-    count = zend_hash_num_elements(seeds);
-
-    /* Allocate our return value and map */
-    z_seeds = ecalloc(count, sizeof(zval*));
-    map = emalloc(sizeof(int)*count);
-
-    /* Fill in and shuffle our map */
-    for (i = 0; i < count; i++) map[i] = i;
-    fyshuffle(map, count);
-
-    /* Iterate over our source array and use our map to create a random list */
-    ZEND_HASH_FOREACH_VAL(seeds, z_ele) {
-        z_seeds[map[index++]] = z_ele;
-    } ZEND_HASH_FOREACH_END();
-
-    efree(map);
-
-    *len = count;
-    return z_seeds;
-}
-
 static void cluster_free_cached_master(redisCachedMaster *cm) {
     size_t i;
 
@@ -1011,7 +966,6 @@ PHP_REDIS_API void cluster_cache_free(redisCachedCluster *rcc) {
         cluster_free_cached_master(&rcc->master[i]);
     }
 
-    /* Free hash key */
     zend_string_release(rcc->hash);
     pefree(rcc->master, 1);
     pefree(rcc, 1);
@@ -1032,23 +986,25 @@ void cluster_init_cache(redisCluster *c, redisCachedCluster *cc) {
     for (i = 0; i < cc->count; i++) map[i] = i;
     fyshuffle(map, cc->count);
 
+    /* Duplicate the hash key so we can invalidate when redirected */
+    c->cache_key = zend_string_copy(cc->hash);
+
     /* Iterate over masters */
     for (i = 0; i < cc->count; i++) {
-        /* Attach cache key */
-        c->cache_key = cc->hash;
-
         /* Grab the next master */
         cm = &cc->master[map[i]];
 
         /* Hash our host and port */
-        keylen = snprintf(key, sizeof(key), "%s:%u", ZSTR_VAL(cm->host.addr),
-                          cm->host.port);
+        keylen = snprintf(key, sizeof(key), "%s:%u", ZSTR_VAL(cm->host.addr), cm->host.port);
 
         /* Create socket */
         sock = redis_sock_create(ZSTR_VAL(cm->host.addr), ZSTR_LEN(cm->host.addr), cm->host.port,
-                                 c->timeout, c->read_timeout, c->persistent,
+                                 c->flags->timeout, c->flags->read_timeout, c->flags->persistent,
                                  NULL, 0);
 
+        /* Stream context */
+        sock->stream_ctx = c->flags->stream_ctx;
+
         /* Add to seed nodes */
         zend_hash_str_update_ptr(c->seeds, key, keylen, sock);
 
@@ -1076,56 +1032,47 @@ void cluster_init_cache(redisCluster *c, redisCachedCluster *cc) {
     efree(map);
 }
 
-/* Initialize seeds */
-PHP_REDIS_API int
-cluster_init_seeds(redisCluster *cluster, HashTable *ht_seeds) {
-    RedisSock *redis_sock;
-    char *str, *psep, key[1024];
-    int key_len, count, i;
-    zval **z_seeds, *z_seed;
+/* Initialize seeds.  By the time we get here we've already validated our
+ * seeds array and know we have a non-empty array of strings all in
+ * host:port format. */
+PHP_REDIS_API void
+cluster_init_seeds(redisCluster *c, zend_string **seeds, uint32_t nseeds)
+{
+    RedisSock *sock;
+    char *seed, *sep, key[1024];
+    int key_len, i, *map;
 
-    /* Get our seeds in a randomized array */
-    z_seeds = cluster_shuffle_seeds(ht_seeds, &count);
+    /* Get a randomized order to hit our seeds */
+    map = ecalloc(nseeds, sizeof(*map));
+    for (i = 0; i < nseeds; i++) map[i] = i;
+    fyshuffle(map, nseeds);
 
-    // Iterate our seeds array
-    for (i = 0; i < count; i++) {
-        if ((z_seed = z_seeds[i]) == NULL) continue;
+    for (i = 0; i < nseeds; i++) {
+        seed = ZSTR_VAL(seeds[map[i]]);
 
-        ZVAL_DEREF(z_seed);
-
-        /* Has to be a string */
-        if (Z_TYPE_P(z_seed) != IS_STRING) continue;
-
-        // Grab a copy of the string
-        str = Z_STRVAL_P(z_seed);
-
-        /* Make sure we have a colon for host:port.  Search right to left in the
-         * case of IPv6 */
-        if ((psep = strrchr(str, ':')) == NULL)
-            continue;
+        sep = strrchr(seed, ':');
+        ZEND_ASSERT(sep != NULL);
 
         // Allocate a structure for this seed
-        redis_sock = redis_sock_create(str, psep-str,
-            (unsigned short)atoi(psep+1), cluster->timeout,
-            cluster->read_timeout, cluster->persistent, NULL, 0);
+        sock = redis_sock_create(seed, sep - seed, atoi(sep + 1),
+                                 c->flags->timeout, c->flags->read_timeout,
+                                 c->flags->persistent, NULL, 0);
 
-        // Set auth information if specified
-        if (cluster->auth) {
-            redis_sock->auth = zend_string_copy(cluster->auth);
-        }
+        /* Stream context */
+        sock->stream_ctx = c->flags->stream_ctx;
+
+        /* Credentials */
+        redis_sock_set_auth(sock, c->flags->user, c->flags->pass);
 
         // Index this seed by host/port
-        key_len = snprintf(key, sizeof(key), "%s:%u", ZSTR_VAL(redis_sock->host),
-            redis_sock->port);
+        key_len = snprintf(key, sizeof(key), "%s:%u", ZSTR_VAL(sock->host),
+            sock->port);
 
         // Add to our seed HashTable
-        zend_hash_str_update_ptr(cluster->seeds, key, key_len, redis_sock);
+        zend_hash_str_update_ptr(c->seeds, key, key_len, sock);
     }
 
-    efree(z_seeds);
-
-    // Success if at least one seed seems valid
-    return zend_hash_num_elements(cluster->seeds) > 0 ? SUCCESS : FAILURE;
+    efree(map);
 }
 
 /* Initial mapping of our cluster keyspace */
@@ -1137,7 +1084,7 @@ PHP_REDIS_API int cluster_map_keyspace(redisCluster *c) {
     // Iterate over seeds until we can get slots
     ZEND_HASH_FOREACH_PTR(c->seeds, seed) {
         // Attempt to connect to this seed node
-        if (seed == NULL || cluster_sock_open(seed) != 0) {
+        if (seed == NULL || redis_sock_server_open(seed) != SUCCESS) {
             continue;
         }
 
@@ -1160,7 +1107,7 @@ PHP_REDIS_API int cluster_map_keyspace(redisCluster *c) {
     // Throw an exception if we couldn't map
     if (!mapped) {
         CLUSTER_THROW_EXCEPTION("Couldn't map cluster keyspace using any provided seed", 0);
-        return -1;
+        return FAILURE;
     }
 
     return SUCCESS;
@@ -1205,8 +1152,7 @@ static int cluster_set_redirection(redisCluster* c, char *msg, int moved)
  *
  * This function will return -1 on a critical error (e.g. parse/communication
  * error, 0 if no redirection was encountered, and 1 if the data was moved. */
-static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type
-                                 )
+static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type)
 {
     size_t sz;
 
@@ -1231,15 +1177,17 @@ static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type
         }
 
         // Check for MOVED or ASK redirection
-        if ((moved = IS_MOVED(inbuf)) || IS_ASK(inbuf)) { // Set our redirection information
-            /* We'll want to invalidate slot cache if we're using one */
-            c->redirections++;
+        if ((moved = IS_MOVED(inbuf)) || IS_ASK(inbuf)) {
+            /* The Redis Cluster specification suggests clients do not update
+             * their slot mapping for an ASK redirection, only for MOVED */
+            if (moved) c->redirections++;
 
+            /* Make sure we can parse the redirection host and port */
             if (cluster_set_redirection(c,inbuf,moved) < 0) {
                 return -1;
             }
 
-            // Data moved
+            /* We've been redirected */
             return 1;
         } else {
             // Capture the error string Redis returned
@@ -1255,7 +1203,7 @@ static int cluster_check_response(redisCluster *c, REDIS_REPLY_TYPE *reply_type
         return -1;
     }
 
-    // For replies that will give us a numberic length, convert it
+    // For replies that will give us a numeric length, convert it
     if (*reply_type != TYPE_LINE) {
         c->reply_len = strtol(c->line_reply, NULL, 10);
     } else {
@@ -1379,8 +1327,7 @@ static int cluster_sock_write(redisCluster *c, const char *cmd, size_t sz,
     /* If in ASK redirection, get/create the node for that host:port, otherwise
      * just use the command socket. */
     if (c->redir_type == REDIR_ASK) {
-        redis_sock = cluster_get_asking_sock(c);
-        if (cluster_send_asking(redis_sock) < 0) {
+        if (cluster_send_asking(c->cmd_sock) < 0) {
             return -1;
         }
     }
@@ -1592,7 +1539,7 @@ PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char
     msstart = mstime();
 
     /* Our main cluster request/reply loop.  This loop runs until we're able to
-     * get a valid reply from a node, hit our "request" timeout, or enounter a
+     * get a valid reply from a node, hit our "request" timeout, or encounter a
      * CLUSTERDOWN state from Redis Cluster. */
     do {
         /* Send MULTI to the socket if we're in MULTI mode but haven't yet */
@@ -1626,10 +1573,13 @@ PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char
                return -1;
            }
 
-           /* Update mapping if the data has MOVED */
            if (c->redir_type == REDIR_MOVED) {
+               /* For MOVED redirection we want to update our cached mapping */
                cluster_update_slot(c);
                c->cmd_sock = SLOT_SOCK(c, slot);
+           } else if (c->redir_type == REDIR_ASK) {
+               /* For ASK redirection we want to redirect but not update slot mapping */
+               c->cmd_sock = cluster_get_asking_sock(c);
            }
         }
 
@@ -1646,11 +1596,9 @@ PHP_REDIS_API short cluster_send_command(redisCluster *c, short slot, const char
         redis_sock_disconnect(c->cmd_sock, 1);
 
         if (timedout) {
-            CLUSTER_THROW_EXCEPTION(
-                "Timed out attempting to find data in the correct node!", 0);
+            CLUSTER_THROW_EXCEPTION("Timed out attempting to find data in the correct node!", 0);
         } else {
-            CLUSTER_THROW_EXCEPTION(
-                "Error processing response from Redis node!", 0);
+            CLUSTER_THROW_EXCEPTION("Error processing response from Redis node!", 0);
         }
 
         return -1;
@@ -1872,7 +1820,6 @@ PHP_REDIS_API void cluster_sub_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *
     zval z_ret, z_args[4];
     sctx->cb.retval = &z_ret;
     sctx->cb.params = z_args;
-    sctx->cb.no_separation = 0;
 
     /* We're in a subscribe loop */
     c->subscribed_slot = c->cmd_slot;
@@ -1994,10 +1941,11 @@ PHP_REDIS_API void cluster_unsub_resp(INTERNAL_FUNCTION_PARAMETERS,
 }
 
 /* Recursive MULTI BULK -> PHP style response handling */
-static void cluster_mbulk_variant_resp(clusterReply *r, zval *z_ret)
+static void cluster_mbulk_variant_resp(clusterReply *r, int null_mbulk_as_null,
+                                       zval *z_ret)
 {
     zval z_sub_ele;
-    int i;
+    long long i;
 
     switch(r->type) {
         case TYPE_INT:
@@ -2018,11 +1966,15 @@ static void cluster_mbulk_variant_resp(clusterReply *r, zval *z_ret)
             }
             break;
         case TYPE_MULTIBULK:
-            array_init(&z_sub_ele);
-            for (i = 0; i < r->elements; i++) {
-                cluster_mbulk_variant_resp(r->element[i], &z_sub_ele);
+            if (r->elements < 0 && null_mbulk_as_null) {
+                add_next_index_null(z_ret);
+            } else {
+                array_init(&z_sub_ele);
+                for (i = 0; i < r->elements; i++) {
+                    cluster_mbulk_variant_resp(r->element[i], null_mbulk_as_null, &z_sub_ele);
+                }
+                add_next_index_zval(z_ret, &z_sub_ele);
             }
-            add_next_index_zval(z_ret, &z_sub_ele);
             break;
         default:
             add_next_index_bool(z_ret, 0);
@@ -2038,14 +1990,14 @@ cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
 {
     clusterReply *r;
     zval zv, *z_arr = &zv;
-    int i;
+    long long i;
 
     // Make sure we can read it
     if ((r = cluster_read_resp(c, status_strings)) == NULL) {
         CLUSTER_RETURN_FALSE(c);
     }
 
-    // Handle ATOMIC vs. MULTI mode in a seperate switch
+    // Handle ATOMIC vs. MULTI mode in a separate switch
     if (CLUSTER_IS_ATOMIC(c)) {
         switch(r->type) {
             case TYPE_INT:
@@ -2069,12 +2021,15 @@ cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
                 }
                 break;
             case TYPE_MULTIBULK:
-                array_init(z_arr);
-
-                for (i = 0; i < r->elements; i++) {
-                    cluster_mbulk_variant_resp(r->element[i], z_arr);
+                if (r->elements < 0 && c->flags->null_mbulk_as_null) {
+                    RETVAL_NULL();
+                } else {
+                    array_init(z_arr);
+                    for (i = 0; i < r->elements; i++) {
+                        cluster_mbulk_variant_resp(r->element[i], c->flags->null_mbulk_as_null, z_arr);
+                    }
+                    RETVAL_ZVAL(z_arr, 0, 0);
                 }
-                RETVAL_ZVAL(z_arr, 0, 0);
                 break;
             default:
                 RETVAL_FALSE;
@@ -2103,7 +2058,11 @@ cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
                 }
                 break;
             case TYPE_MULTIBULK:
-                cluster_mbulk_variant_resp(r, &c->multi_resp);
+                if (r->elements < 0 && c->flags->null_mbulk_as_null) {
+                    add_next_index_null(&c->multi_resp);
+                } else {
+                    cluster_mbulk_variant_resp(r, c->flags->null_mbulk_as_null, &c->multi_resp);
+                }
                 break;
             default:
                 add_next_index_bool(&c->multi_resp, 0);
@@ -2124,7 +2083,8 @@ PHP_REDIS_API void cluster_variant_resp(INTERNAL_FUNCTION_PARAMETERS, redisClust
 PHP_REDIS_API void cluster_variant_raw_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
                                             void *ctx)
 {
-    cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, c->flags->reply_literal, ctx);
+    cluster_variant_resp_generic(INTERNAL_FUNCTION_PARAM_PASSTHRU, c,
+                                 c->flags->reply_literal, ctx);
 }
 
 PHP_REDIS_API void cluster_variant_resp_strings(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c,
@@ -2139,23 +2099,25 @@ PHP_REDIS_API void cluster_gen_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
 {
     zval z_result;
 
-    /* Return FALSE if we didn't get a multi-bulk response */
-    if (c->reply_type != TYPE_MULTIBULK) {
+    /* Abort if the reply isn't MULTIBULK or has an invalid length */
+    if (c->reply_type != TYPE_MULTIBULK || c->reply_len < -1) {
         CLUSTER_RETURN_FALSE(c);
     }
 
-    /* Allocate our array */
-    array_init(&z_result);
+    if (c->reply_len == -1 && c->flags->null_mbulk_as_null) {
+        ZVAL_NULL(&z_result);
+    } else {
+        array_init(&z_result);
 
-    /* Consume replies as long as there are more than zero */
-    if (c->reply_len > 0) {
-        /* Push serialization settings from the cluster into our socket */
-        c->cmd_sock->serializer = c->flags->serializer;
+        if (c->reply_len > 0) {
+            /* Push serialization settings from the cluster into our socket */
+            c->cmd_sock->serializer = c->flags->serializer;
 
-        /* Call our specified callback */
-        if (cb(c->cmd_sock, &z_result, c->reply_len, ctx) == FAILURE) {
-            zval_dtor(&z_result);
-            CLUSTER_RETURN_FALSE(c);
+            /* Call our specified callback */
+            if (cb(c->cmd_sock, &z_result, c->reply_len, ctx) == FAILURE) {
+                zval_dtor(&z_result);
+                CLUSTER_RETURN_FALSE(c);
+            }
         }
     }
 
@@ -2300,14 +2262,17 @@ PHP_REDIS_API void
 cluster_xread_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
     zval z_streams;
 
-    array_init(&z_streams);
-
     c->cmd_sock->serializer = c->flags->serializer;
     c->cmd_sock->compression = c->flags->compression;
 
-    if (redis_read_stream_messages_multi(c->cmd_sock, c->reply_len, &z_streams) < 0) {
-        zval_dtor(&z_streams);
-        CLUSTER_RETURN_FALSE(c);
+    if (c->reply_len == -1 && c->flags->null_mbulk_as_null) {
+        ZVAL_NULL(&z_streams);
+    } else {
+        array_init(&z_streams);
+        if (redis_read_stream_messages_multi(c->cmd_sock, c->reply_len, &z_streams) < 0) {
+            zval_dtor(&z_streams);
+            CLUSTER_RETURN_FALSE(c);
+        }
     }
 
     if (CLUSTER_IS_ATOMIC(c)) {
@@ -2355,11 +2320,40 @@ cluster_xinfo_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx)
     add_next_index_zval(&c->multi_resp, &z_ret);
 }
 
+static void
+cluster_acl_custom_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx,
+                        int (*cb)(RedisSock*, zval*, long))
+{
+    zval z_ret;
+
+    array_init(&z_ret);
+    if (cb(c->cmd_sock, &z_ret, c->reply_len) != SUCCESS) {
+        zval_dtor(&z_ret);
+        CLUSTER_RETURN_FALSE(c);
+    }
+
+    if (CLUSTER_IS_ATOMIC(c)) {
+        RETURN_ZVAL(&z_ret, 0, 1);
+    }
+    add_next_index_zval(&c->multi_resp, &z_ret);
+}
+
+PHP_REDIS_API void
+cluster_acl_getuser_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    cluster_acl_custom_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, ctx, redis_read_acl_getuser_reply);
+}
+
+PHP_REDIS_API void
+cluster_acl_log_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluster *c, void *ctx) {
+    cluster_acl_custom_resp(INTERNAL_FUNCTION_PARAM_PASSTHRU, c, ctx, redis_read_acl_log_reply);
+}
+
 /* MULTI BULK response loop where we might pull the next one */
 PHP_REDIS_API zval *cluster_zval_mbulk_resp(INTERNAL_FUNCTION_PARAMETERS,
                                      redisCluster *c, int pull, mbulk_cb cb, zval *z_ret)
 {
     ZVAL_NULL(z_ret);
+
     // Pull our next response if directed
     if (pull) {
         if (cluster_check_response(c, &c->reply_type) < 0)
@@ -2433,7 +2427,7 @@ PHP_REDIS_API void cluster_mbulk_mget_resp(INTERNAL_FUNCTION_PARAMETERS,
         mbulk_resp_loop(c->cmd_sock, mctx->z_multi, c->reply_len, NULL) == FAILURE;
 
     // If we had a failure, pad results with FALSE to indicate failure.  Non
-    // existant keys (e.g. for MGET will come back as NULL)
+    // existent keys (e.g. for MGET will come back as NULL)
     if (fail) {
         while (mctx->count--) {
             add_next_index_bool(mctx->z_multi, 0);
@@ -2447,6 +2441,8 @@ PHP_REDIS_API void cluster_mbulk_mget_resp(INTERNAL_FUNCTION_PARAMETERS,
         } else {
             add_next_index_zval(&c->multi_resp, mctx->z_multi);
         }
+
+        efree(mctx->z_multi);
     }
 
     // Clean up this context item
@@ -2478,10 +2474,11 @@ PHP_REDIS_API void cluster_msetnx_resp(INTERNAL_FUNCTION_PARAMETERS, redisCluste
     // Set return value if it's our last response
     if (mctx->last) {
         if (CLUSTER_IS_ATOMIC(c)) {
-            RETVAL_ZVAL(mctx->z_multi, 0, 1);
+            RETVAL_ZVAL(mctx->z_multi, 0, 0);
         } else {
             add_next_index_zval(&c->multi_resp, mctx->z_multi);
         }
+        efree(mctx->z_multi);
     }
 
     // Free multi context
@@ -2638,7 +2635,6 @@ int mbulk_resp_loop(RedisSock *redis_sock, zval *z_result,
             }
             efree(line);
         } else {
-            if (line) efree(line);
             add_next_index_bool(z_result, 0);
         }
     }
@@ -2654,7 +2650,7 @@ int mbulk_resp_loop_zipstr(RedisSock *redis_sock, zval *z_result,
     int line_len, key_len = 0;
     long long idx = 0;
 
-    // Our count wil need to be divisible by 2
+    // Our count will need to be divisible by 2
     if (count % 2 != 0) {
         return -1;
     }
@@ -2732,11 +2728,11 @@ int mbulk_resp_loop_assoc(RedisSock *redis_sock, zval *z_result,
                           long long count, void *ctx)
 {
     char *line;
-    int line_len,i = 0;
+    int line_len, i;
     zval *z_keys = ctx;
 
     // Loop while we've got replies
-    while (count--) {
+    for (i = 0; i < count; ++i) {
         zend_string *zstr = zval_get_string(&z_keys[i]);
         line = redis_sock_read(redis_sock, &line_len);
 
@@ -2755,9 +2751,6 @@ int mbulk_resp_loop_assoc(RedisSock *redis_sock, zval *z_result,
         // Clean up key context
         zend_string_release(zstr);
         zval_dtor(&z_keys[i]);
-
-        // Move to the next key
-        i++;
     }
 
     // Clean up our keys overall
@@ -2767,4 +2760,169 @@ int mbulk_resp_loop_assoc(RedisSock *redis_sock, zval *z_result,
     return SUCCESS;
 }
 
+/* Free an array of zend_string seeds */
+void free_seed_array(zend_string **seeds, uint32_t nseeds) {
+    int i;
+
+    if (seeds == NULL)
+        return;
+
+    for (i = 0; i < nseeds; i++)
+        zend_string_release(seeds[i]);
+
+    efree(seeds);
+}
+
+static zend_string **get_valid_seeds(HashTable *input, uint32_t *nseeds) {
+    HashTable *valid;
+    uint32_t count, idx = 0;
+    zval *z_seed;
+    zend_string *zkey, **seeds = NULL;
+
+    /* Short circuit if we don't have any sees */
+    count = zend_hash_num_elements(input);
+    if (count == 0)
+        return NULL;
+
+    ALLOC_HASHTABLE(valid);
+    zend_hash_init(valid, count, NULL, NULL, 0);
+
+    ZEND_HASH_FOREACH_VAL(input, z_seed) {
+        ZVAL_DEREF(z_seed);
+
+        if (Z_TYPE_P(z_seed) != IS_STRING) {
+            php_error_docref(NULL, E_WARNING, "Skipping non-string entry in seeds array");
+            continue;
+        } else if (strrchr(Z_STRVAL_P(z_seed), ':') == NULL) {
+            php_error_docref(NULL, E_WARNING,
+                "Seed '%s' not in host:port format, ignoring", Z_STRVAL_P(z_seed));
+            continue;
+        }
+
+        /* Add as a key to avoid duplicates */
+        zend_hash_str_update_ptr(valid, Z_STRVAL_P(z_seed), Z_STRLEN_P(z_seed), NULL);
+    } ZEND_HASH_FOREACH_END();
+
+    /* We need at least one valid seed */
+    count = zend_hash_num_elements(valid);
+    if (count == 0)
+        goto cleanup;
+
+    /* Populate our return array */
+    seeds = ecalloc(count, sizeof(*seeds));
+    ZEND_HASH_FOREACH_STR_KEY(valid, zkey) {
+        seeds[idx++] = zend_string_copy(zkey);
+    } ZEND_HASH_FOREACH_END();
+
+    *nseeds = idx;
+
+cleanup:
+    zend_hash_destroy(valid);
+    FREE_HASHTABLE(valid);
+
+    return seeds;
+}
+
+/* Validate cluster construction arguments and return a sanitized and validated
+ * array of seeds */
+zend_string**
+cluster_validate_args(double timeout, double read_timeout, HashTable *seeds,
+                      uint32_t *nseeds, char **errstr)
+{
+    zend_string **retval;
+
+    if (timeout < 0L || timeout > INT_MAX) {
+        if (errstr) *errstr = "Invalid timeout";
+        return NULL;
+    }
+
+    if (read_timeout < 0L || read_timeout > INT_MAX) {
+        if (errstr) *errstr = "Invalid read timeout";
+        return NULL;
+    }
+
+    retval = get_valid_seeds(seeds, nseeds);
+    if (retval == NULL && errstr)
+        *errstr = "No valid seeds detected";
+
+    return retval;
+}
+
+/* Helper function to compare to host:port seeds */
+static int cluster_cmp_seeds(const void *a, const void *b) {
+    zend_string *za = *(zend_string **)a;
+    zend_string *zb = *(zend_string **)b;
+    return strcmp(ZSTR_VAL(za), ZSTR_VAL(zb));
+}
+
+static void cluster_swap_seeds(void *a, void *b) {
+    zend_string **za, **zb, *tmp;
+
+    za = a;
+    zb = b;
+
+    tmp = *za;
+    *za = *zb;
+    *zb = tmp;
+}
+
+/* Turn an array of cluster seeds into a string we can cache.  If we get here we know
+ * we have at least one entry and that every entry is a string in the form host:port */
+#define SLOT_CACHE_PREFIX "phpredis_slots:"
+zend_string *cluster_hash_seeds(zend_string **seeds, uint32_t count) {
+    smart_str hash = {0};
+    size_t i;
+
+    /* Sort our seeds so any any array with identical seeds hashes to the same key
+     * regardless of what order the user gives them to us in. */
+    zend_sort(seeds, count, sizeof(*seeds), cluster_cmp_seeds, cluster_swap_seeds);
+
+    /* Global phpredis hash prefix */
+    smart_str_appendl(&hash, SLOT_CACHE_PREFIX, sizeof(SLOT_CACHE_PREFIX) - 1);
+
+    /* Construct our actual hash */
+    for (i = 0; i < count; i++) {
+        smart_str_appendc(&hash, '[');
+        smart_str_append_ex(&hash, seeds[i], 0);
+        smart_str_appendc(&hash, ']');
+    }
+
+    /* Null terminate */
+    smart_str_0(&hash);
+
+    /* Return the internal zend_string */
+    return hash.s;
+}
+
+PHP_REDIS_API redisCachedCluster *cluster_cache_load(zend_string *hash) {
+    zend_resource *le;
+
+    /* Look for cached slot information */
+    le = zend_hash_find_ptr(&EG(persistent_list), hash);
+
+    if (le != NULL) {
+        /* Sanity check on our list type */
+        if (le->type != le_cluster_slot_cache) {
+            php_error_docref(0, E_WARNING, "Invalid slot cache resource");
+            return NULL;
+        }
+
+        /* Success, return the cached entry */
+        return le->ptr;
+    }
+
+    /* Not found */
+    return NULL;
+}
+
+/* Cache a cluster's slot information in persistent_list if it's enabled */
+PHP_REDIS_API int cluster_cache_store(zend_string *hash, HashTable *nodes) {
+    redisCachedCluster *cc = cluster_cache_create(hash, nodes);
+
+    redis_register_persistent_resource(cc->hash, cc, le_cluster_slot_cache);
+
+    return SUCCESS;
+}
+
+
 /* vim: set tabstop=4 softtabstop=4 expandtab shiftwidth=4: */
