@@ -46,6 +46,8 @@ class Redis_Cluster_Test extends Redis_Test {
     public function testDoublePipeNoOp() { return $this->markTestSkipped(); }
     public function testSwapDB() { return $this->markTestSkipped(); }
     public function testConnectException() { return $this->markTestSkipped(); }
+    public function testTlsConnect() { return $this->markTestSkipped(); }
+    public function testInvalidAuthArgs() { return $this->markTestSkipped(); }
 
     /* Session locking feature is currently not supported in in context of Redis Cluster.
        The biggest issue for this is the distribution nature of Redis cluster */
@@ -62,7 +64,9 @@ class Redis_Cluster_Test extends Redis_Test {
     public function testSession_lockWaitTime() { return $this->markTestSkipped(); }
 
     /* Load our seeds on construction */
-    public function __construct() {
+    public function __construct($str_host, $i_port, $str_auth) {
+        parent::__construct($str_host, $i_port, $str_auth);
+
         $str_nodemap_file = dirname($_SERVER['PHP_SELF']) . '/nodes/nodemap';
 
         if (!file_exists($str_nodemap_file)) {
@@ -87,7 +91,7 @@ class Redis_Cluster_Test extends Redis_Test {
 
     /* Override newInstance as we want a RedisCluster object */
     protected function newInstance() {
-        return new RedisCluster(NULL, self::$_arr_node_map);
+        return new RedisCluster(NULL, self::$_arr_node_map, 30, 30, true, $this->getAuth());
     }
 
     /* Overrides for RedisTest where the function signature is different.  This
@@ -220,6 +224,55 @@ class Redis_Cluster_Test extends Redis_Test {
         $this->assertEquals($i_scan_count, $i_key_count);
     }
 
+    public function testScanPrefix() {
+        $arr_prefixes = ['prefix-a:', 'prefix-b:'];
+        $str_id = uniqid();
+
+        $arr_keys = [];
+        foreach ($arr_prefixes as $str_prefix) {
+            $this->redis->setOption(Redis::OPT_PREFIX, $str_prefix);
+            $this->redis->set($str_id, "LOLWUT");
+            $arr_keys[$str_prefix] = $str_id;
+        }
+
+        $this->redis->setOption(Redis::OPT_SCAN, Redis::SCAN_RETRY);
+        $this->redis->setOption(Redis::OPT_SCAN, Redis::SCAN_PREFIX);
+
+        foreach ($arr_prefixes as $str_prefix) {
+            $arr_prefix_keys = [];
+            $this->redis->setOption(Redis::OPT_PREFIX, $str_prefix);
+
+            foreach ($this->redis->_masters() as $arr_master) {
+                $it = NULL;
+                while ($arr_iter = $this->redis->scan($it, $arr_master, "*$str_id*")) {
+                    foreach ($arr_iter as $str_key) {
+                        $arr_prefix_keys[$str_prefix] = $str_key;
+                    }
+                }
+            }
+
+            $this->assertTrue(count($arr_prefix_keys) == 1 && isset($arr_prefix_keys[$str_prefix]));
+        }
+
+        $this->redis->setOption(Redis::OPT_SCAN, Redis::SCAN_NOPREFIX);
+
+        $arr_scan_keys = [];
+
+        foreach ($this->redis->_masters() as $arr_master) {
+            $it = NULL;
+            while ($arr_iter = $this->redis->scan($it, $arr_master, "*$str_id*")) {
+                foreach ($arr_iter as $str_key) {
+                    $arr_scan_keys[] = $str_key;
+                }
+            }
+        }
+
+        /* We should now have both prefixs' keys */
+        foreach ($arr_keys as $str_prefix => $str_id) {
+            $this->assertTrue(in_array("${str_prefix}${str_id}", $arr_scan_keys));
+        }
+    }
+
     // Run some simple tests against the PUBSUB command.  This is problematic, as we
     // can't be sure what's going on in the instance, but we can do some things.
     public function testPubSub() {
@@ -636,12 +689,19 @@ class Redis_Cluster_Test extends Redis_Test {
         $this->redis->setOption(Redis::OPT_REPLY_LITERAL, false);
     }
 
+    /* Redis and RedisCluster use the same handler for the ACL command but verify we can direct
+       the command to a specific node. */
+    public function testAcl() {
+        if ( ! $this->minVersionCheck("6.0"))
+            return $this->markTestSkipped();
+
+        $this->assertInArray('default', $this->redis->acl('foo', 'USERS'));
+    }
+
     public function testSession()
     {
         @ini_set('session.save_handler', 'rediscluster');
-        @ini_set('session.save_path', implode('&', array_map(function ($seed) {
-            return 'seed[]=' . $seed;
-        }, self::$_arr_node_map)) . '&failover=error');
+        @ini_set('session.save_path', $this->getFullHostPath() . '&failover=error');
         if (!@session_start()) {
             return $this->markTestSkipped();
         }
@@ -656,7 +716,7 @@ class Redis_Cluster_Test extends Redis_Test {
 
         $pong = 0;
         for ($i = 0; $i < 10; $i++) {
-            $obj_rc = new RedisCluster(NULL, self::$_arr_node_map);
+            $obj_rc = new RedisCluster(NULL, self::$_arr_node_map, 30, 30, true, $this->getAuth());
             $pong += $obj_rc->ping("key:$i");
         }
 
@@ -665,16 +725,51 @@ class Redis_Cluster_Test extends Redis_Test {
         ini_set('redis.clusters.cache_slots', 0);
     }
 
+    /* Regression test for connection pool liveness checks */
+    public function testConnectionPool() {
+        $prev_value = ini_get('redis.pconnect.pooling_enabled');
+        ini_set('redis.pconnect.pooling_enabled', 1);
+
+        $pong = 0;
+        for ($i = 0; $i < 10; $i++) {
+            $obj_rc = new RedisCluster(NULL, self::$_arr_node_map, 30, 30, true, $this->getAuth());
+            $pong += $obj_rc->ping("key:$i");
+        }
+
+        $this->assertEquals($pong, $i);
+        ini_set('redis.pconnect.pooling_enabled', $prev_value);
+    }
+
     /**
      * @inheritdoc
      */
     protected function getFullHostPath()
     {
-        $hosts = array_map(function ($host) {
-            return 'seed[]=' . $host . '';
-        }, self::$_arr_node_map);
+        $auth = $this->getAuthFragment();
 
-        return implode('&', $hosts);
+        return implode('&', array_map(function ($host) {
+            return 'seed[]=' . $host;
+        }, self::$_arr_node_map)) . ($auth ? "&$auth" : '');
+    }
+
+    /* Test correct handling of null multibulk replies */
+    public function testNullArray() {
+        $key = "key:arr";
+        $this->redis->del($key);
+
+        foreach ([false => [], true => NULL] as $opt => $test) {
+            $this->redis->setOption(Redis::OPT_NULL_MULTIBULK_AS_NULL, $opt);
+
+            $r = $this->redis->rawCommand($key, "BLPOP", $key, .05);
+            $this->assertEquals($test, $r);
+
+            $this->redis->multi();
+            $this->redis->rawCommand($key, "BLPOP", $key, .05);
+            $r = $this->redis->exec();
+            $this->assertEquals([$test], $r);
+        }
+
+        $this->redis->setOption(Redis::OPT_NULL_MULTIBULK_AS_NULL, false);
     }
 }
 ?>
